/*
 *  DblList.h by Davide Libenzi (Macro based doublelink list implementation)
 *  Copyright (C) 2008  Davide Libenzi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */

#ifndef _DBLLIST_H
#define _DBLLIST_H

struct DblListHead {
	struct DblListHead *next;
	struct DblListHead *prev;
};

#define DBLLISTHEAD_INIT(name) { &(name), &(name) }

#define DBLLISTHEAD(name) struct DblListHead name = DBLLISTHEAD_INIT(name)

#define DBLL_INIT_LIST_HEAD(ptr)				\
	do {							\
		(ptr)->next = (ptr); (ptr)->prev = (ptr);	\
	} while (0)

#define INIT_LIST_LINK(ptr)				\
	do {						\
		(ptr)->next = NULL; (ptr)->prev = NULL; \
	} while (0)

#define DBLL_LIST_ADD(new, previ, nexti)		\
	do {						\
		struct DblListHead *prev = previ;	\
		struct DblListHead *next = nexti;	\
		next->prev = new;			\
		(new)->next = next;			\
		(new)->prev = prev;			\
		prev->next = new;			\
	} while (0)

#define DBLL_LIST_ADDH(new, head) DBLL_LIST_ADD(new, head, (head)->next)
#define DBLL_LIST_ADDT(new, head) DBLL_LIST_ADD(new, (head)->prev, head)

#define DBLL_LIST_UNLINK(previ, nexti)		\
	do {					\
		(nexti)->prev = previ;		\
		(previ)->next = nexti;		\
	} while (0)

#define DBLL_LIST_DEL(entry)					\
	do {							\
		DBLL_LIST_UNLINK((entry)->prev, (entry)->next); \
		DBLL_INIT_LIST_HEAD(entry);			\
	} while (0)

#define DBLL_LIST_EMPTY(head) ((head)->next == head)

#define DBLL_LIST_SPLICE(list, head)					\
	do {								\
		struct DblListHead *first = (list)->next;		\
		if (first != list) {					\
			struct DblListHead *last = (list)->prev;	\
			struct DblListHead *at = (head)->next;		\
			(first)->prev = head;				\
			(head)->next = first;				\
			(last)->next = at;				\
			(at)->prev = last;				\
		}							\
	} while (0)

#define DBLL_LIST_COUNT(cnt, pos, head)					\
	for ((cnt) = 0, (pos) = (head)->next; (pos) != (head); (cnt)++, (pos) = (pos)->next)

#define DBLL_LIST_ENTRY(ptr, type, member) ((type *)((char *)(ptr) - (unsigned long)(&((type *)0)->member)))
#define DBLL_LIST_OFFENTRY(ptr, type, off) ((type *)((char *)(ptr) - (off)))
#define DBLL_LIST_FOR_EACH(pos, head) for (pos = (head)->next; pos != (head); pos = (pos)->next)
#define DBLL_LIST_FIRST(head) ((head)->next != (head) ? (head)->next: NULL)
#define DBLL_LIST_LAST(head) ((head)->prev != (head) ? (head)->prev: NULL)
#define DBLL_LIST_LINKED(ptr) ((ptr)->prev != NULL && (ptr)->next != NULL)
#define DBLL_LIST_NEXT(ptr, head) ((ptr)->next != (head) ? (ptr)->next: NULL)
#define DBLL_LIST_PREV(ptr, head) ((ptr)->prev != (head) ? (ptr)->prev: NULL)


#endif

